### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################


# import Zope3 interfaces
from zope.annotation.interfaces import IAnnotatable
from zope.security.interfaces import IPermission, IPrincipal
from zope.securitypolicy.interfaces import IRole, IPrincipalRoleManager

# import local interfaces
from interfaces import ILocalRoleManager, ISecurityManager, ISecurityManagerRoleChecker, IBaseRoleEvent, IGrantedRoleEvent, IRevokedRoleEvent

# import Zope3 packages
from zope.app import zapi
from zope.component import adapts
import zope.event
from zope.interface import implements
from zope.lifecycleevent import ObjectModifiedEvent
from zope.security.management import checkPermission

# import local packages
from interfaces import SecurityManagerException
from ztfy.utils import request

from ztfy.security import _


class BaseRoleEvent(ObjectModifiedEvent):

    implements(IBaseRoleEvent)

    def __init__(self, object, principal, role):
        super(BaseRoleEvent, self).__init__(object)
        self.principal = principal
        self.role = role


class GrantedRoleEvent(BaseRoleEvent):

    implements(IGrantedRoleEvent)


class RevokedRoleEvent(BaseRoleEvent):

    implements(IRevokedRoleEvent)


class SecurityManagerAdapter(object):

    adapts(IAnnotatable)
    implements(ISecurityManager)

    def __init__(self, context):
        self.context = context

    # ISecurityManagerBase methods
    def canUsePermission(self, permission):
        if IPermission.providedBy(permission):
            permission = IPermission(permission).id
        try:
            return checkPermission(permission, self.context)
        except:
            return False

    def canView(self):
        return self.canUsePermission('zope.View')

    # ISecurityManagerReader methods
    def getLocalRoles(self, principal=None):
        if principal is None:
            principal = request.getPrincipal()
        if not principal:
            return None
        if IPrincipal.providedBy(principal):
            principal = IPrincipal(principal).id
        manager = IPrincipalRoleManager(self.context)
        roles = manager.getRolesForPrincipal(principal)
        return { 'allow': set((role[0] for role in roles if role[1].getName() == 'Allow')),
                 'deny':  set((role[0] for role in roles if role[1].getName() == 'Deny')) }

    def getLocalAllowedRoles(self, principal=None):
        roles = self.getLocalRoles(principal)
        if not roles:
            return None
        return roles['allow']

    def getLocalDeniedRoles(self, principal=None):
        roles = self.getLocalRoles(principal)
        if not roles:
            return None
        return roles['deny']

    def getRoles(self, principal=None):
        result = { 'allow': set(),
                   'deny':  set() }
        context = self.context
        while context is not None:
            roles = ISecurityManager(context).getLocalRoles(principal)
            if roles:
                for role in roles['allow']:
                    if role not in result['deny']:
                        result['allow'].add(role)
                for role in roles['deny']:
                    if role not in result['allow']:
                        result['deny'].add(role)
            context = zapi.getParent(context)
        return result

    def getAllowedRoles(self, principal=None):
        return self.getRoles(principal)['allow']

    def getDeniedRoles(self, principal=None):
        return self.getRoles(principal)['deny']

    def getLocalPrincipals(self, role):
        if IRole.providedBy(role):
            role = IRole(role).id
        manager = IPrincipalRoleManager(self.context)
        principals = manager.getPrincipalsForRole(role)
        return { 'allow': set((principal[0] for principal in principals if principal[1].getName() == 'Allow')),
                 'deny':  set((principal[0] for principal in principals if principal[1].getName() == 'Deny')) }

    def getLocalAllowedPrincipals(self, role):
        return self.getLocalPrincipals(role)['allow']

    def getLocalDeniedPrincipals(self, role):
        return self.getLocalPrincipals(role)['deny']

    def getPrincipals(self, role):
        result = { 'allow': set(),
                   'deny':  set() }
        context = self.context
        while context is not None:
            principals = ISecurityManager(context).getLocalPrincipals(role)
            for principal in principals['allow']:
                if principal not in result['deny']:
                    result['allow'].add(principal)
            for principal in principals['deny']:
                if principal not in result['allow']:
                    result['deny'].add(principal)
            context = zapi.getParent(context)
        return result

    def getAllowedPrincipals(self, role):
        return self.getPrincipals(role)['allow']

    def getDeniedPrincipals(self, role):
        return self.getPrincipals(role)['deny']

    def canUseRole(self, role, principal=None):
        if IRole.providedBy(role):
            role = IRole(role).id
        if principal is None:
            principal = request.getPrincipal()
        if IPrincipal.providedBy(principal):
            principal = principal.id
        if principal == 'zope.manager':
            return True
        return role in self.getAllowedRoles(principal)

    # ISecurityManagerWriter methods
    def _getLocalRoles(self):
        manager = ILocalRoleManager(self.context, None)
        if manager is None:
            return None
        return manager.__roles__

    def grantRole(self, role, principal, notify=True):
        if IRole.providedBy(role):
            role = IRole(role).id
        if IPrincipal.providedBy(principal):
            principal = IPrincipal(principal).id
        roles = self._getLocalRoles()
        if roles and (role not in roles):
            raise SecurityManagerException, _("The role %s is not handled by this object's class") % role
        checker = ISecurityManagerRoleChecker(self.context, None)
        if (checker is not None) and (not checker.canGrantRole(role, principal)):
            raise SecurityManagerException, _("You are not authorized to grant role %s to principal %s") % (role, principal)
        prm = IPrincipalRoleManager(self.context)
        prm.assignRoleToPrincipal(role, principal)
        if notify:
            zope.event.notify(GrantedRoleEvent(self.context, principal, role))

    def unsetRole(self, role, principal, notify=True):
        if IRole.providedBy(role):
            role = IRole(role).id
        if IPrincipal.providedBy(principal):
            principal = IPrincipal(principal).id
        roles = self._getLocalRoles()
        if roles and (role not in roles):
            raise SecurityManagerException, _("The role %s is not handled by this object's class") % role
        checker = ISecurityManagerRoleChecker(self.context, None)
        if (checker is not None) and (not checker.canRevokeRole(role, principal)):
            raise SecurityManagerException, _("You are not authorized to revoke role %s for principal %s") % (role, principal)
        prm = IPrincipalRoleManager(self.context)
        prm.unsetRoleForPrincipal(role, principal)
        if notify:
            zope.event.notify(RevokedRoleEvent(self.context, principal, role))

    def revokeRole(self, role, principal, notify=True):
        if IRole.providedBy(role):
            role = IRole(role).id
        if IPrincipal.providedBy(principal):
            principal = IPrincipal(principal).id
        roles = self._getLocalRoles()
        if roles and (role not in roles):
            raise SecurityManagerException, _("The role %s is not handled by this object's class") % role
        checker = ISecurityManagerRoleChecker(self.context, None)
        if (checker is not None) and (not checker.canRevokeRole(role, principal)):
            raise SecurityManagerException, _("You are not authorized to revoke role %s for principal %s") % (role, principal)
        prm = IPrincipalRoleManager(self.context)
        prm.removeRoleFromPrincipal(role, principal)
        if notify:
            zope.event.notify(RevokedRoleEvent(self.context, principal, role))


def getSecurityManager(context):
    while context is not None:
        sm = ISecurityManager(context, None)
        if sm is not None:
            return sm
        context = zapi.getParent(context)
    return None



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################


# import Zope3 interfaces

# import local interfaces

# import Zope3 packages
from zope.interface import Interface, Attribute
from zope.lifecycleevent.interfaces import IObjectModifiedEvent
from zope.schema import Tuple, TextLine

# import local packages

from ztfy.security import _


class IBaseRoleEvent(IObjectModifiedEvent):
    """Base interface for roles events"""

    principal = Attribute(_("Name of the principal who received the new role"))

    role = Attribute(_("The role who was granted"))


class IGrantedRoleEvent(IBaseRoleEvent):
    """Event interface notified when a role is granted"""


class IRevokedRoleEvent(IBaseRoleEvent):
    """Event interface notified when a role is revoked"""


class ILocalRoleManager(Interface):
    """This interface is used to define classes which can handle local roles"""

    __roles__ = Tuple(title=_("Local roles"),
                      description=_("List of roles IDs handled by local context"),
                      required=True,
                      readonly=True,
                      value_type=TextLine(title=_("Role ID")))


class SecurityManagerException(Exception):
    """Security manager exception"""


class ISecurityManagerRoleChecker(Interface):
    """Security manager role checker"""

    def canGrantRole(role, principal):
        """Return True is role can be granted to principal"""

    def canRevokeRole(role, principal):
        """Return True is role can be revoked from principal"""


class ISecurityManagerBase(Interface):
    """Untrusted security manager interface"""

    def canUsePermission(permission):
        """Return true or false to specify permission usage for given principal"""

    def canView():
        """Return true or false if 'zope.View' permission is granted to given principal"""


class ISecurityManagerReader(Interface):
    """Wrapper interface around Zope3 security roles and permissions"""

    def getLocalRoles(principal=None):
        """Get principal allowed and denied roles on current object

        Result is given as a dictionary :
        { 'allow': ['role1','role2'], 'deny': ['role3',] }
        """

    def getLocalAllowedRoles(principal=None):
        """Get list of locally allowed roles"""

    def getLocalDeniedRoles(principal=None):
        """Get list of locally denied roles"""

    def getRoles(principal=None):
        """Get list of roles, including inherited ones
        
        Result is given as a dictionary :
        { 'allow': ['role1','role2'], 'deny': ['role3',] }
        """

    def getAllowedRoles(principal=None):
        """Get list of allowed roles, including inherited ones"""

    def getDeniedRoles(principal=None):
        """Get list of denied roles, including inherited ones"""

    def getLocalPrincipals(role):
        """Get list of principals with locally defined role
        
        Result is given as a dictionary :
        { 'allow': ['principal1','principal2'], 'deny': ['principal3',] }
        """

    def getLocalAllowedPrincipals(role):
        """Get list of principals with locally granted role"""

    def getLocalDeniedPrincipals(role):
        """Get list of principals with locally denied role"""

    def getPrincipals(role):
        """Get list of principals with access defined for allowed role, including inherited ones
        
        Result is given as a dictionary :
        { 'allow': ['principal1','principal2'], 'deny': ['principal3',] }
        """

    def getAllowedPrincipals(role):
        """Get list of principals with granted access to specified role, including inherited ones"""

    def getDeniedPrincipals(role):
        """Get list of principals with denied access to specified role, including inherited ones"""

    def canUseRole(role, principal=None):
        """Return true or false to specify role usage for given principal"""


class ISecurityManagerWriter(Interface):
    """Wrapper interface around Zope3 security affectations"""

    def grantRole(role, principal, notify=True):
        """Grant given role to the principal"""

    def unsetRole(role, principal, notify=True):
        """Revoke given role to the principal, falling back to normal acquisition"""

    def revokeRole(role, principal, notify=True):
        """Forbid usage of the given role to the principal, ignoring acquisition"""


class ISecurityManager(ISecurityManagerBase, ISecurityManagerReader, ISecurityManagerWriter):
    """Marker interface for security management wrapper"""



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008-2009 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################


# import Zope3 interfaces
from zope.schema.interfaces import ITextLine, IList

# import local interfaces

# import Zope3 packages
from zope.interface import implements
from zope.schema import TextLine, Iterable

# import local packages


class IPrincipal(ITextLine):
    """Marker interface to define principal ID field"""


class IPrincipalList(IList):
    """Marker interface to define a list of principal IDs"""



class Principal(TextLine):
    """Schema field to store a principal ID"""

    implements(IPrincipal)


class PrincipalList(Iterable):
    """Schema field to store a list of principal IDs"""

    implements(IPrincipalList)
